/*
 *  Released under the GNU LGPL.  See http://www.gnu.org/licenses/lgpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  This library is free software * you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundatation * either version 2.1 of the License, or
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY * without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 *  License for more details.
 */

/*	FINDFILE.C	*/
/*
	This file contains the routine to locate a file, utilizing the INCLUDE
	environment variable for the directories to search.

	Interface:

		findfile(envname, filename, target_buf)

		where envname is the envvironment name (like PATH),
		      filename is the name of the file to be searched for,
		      target_buf is the buffer to place the path name if found.

		if the file is found, findfile return 1 and the pathname,
		otherwise it returns 0.

	This program uses the environ routines to access the PATH variable.

	Stack requirements:  ~300 bytes

*/

findfile(envname, filename, target_buf)
	char *envname, *filename, *target_buf; {
	int fid;
	char paths[256], *p_ptr, *t_ptr;

	/* first check in the local directory */
	strcpy(target_buf, filename);
	fid = open(target_buf, 0);
	if (fid >= 0)  {				/* got it */
		close(fid);
		return (1);
		}
	fid = getenv(envname, paths, 256);
	p_ptr = paths;

	while (*p_ptr != 0) {
		/* copy the directory name */
		t_ptr = target_buf;
		while (*p_ptr != ';' && *p_ptr != 0) {
			*t_ptr++ = *p_ptr++;
			}
		if (*(t_ptr-1) != '/' && *(t_ptr-1) != '\\') *t_ptr++ = '\\';
		*t_ptr = 0;
		if (*p_ptr) p_ptr++;		/* beyond the ';' */
		strcat(target_buf, filename);
		fid = open(target_buf, 0);
		if (fid >= 0)  {				/* got it */
			close(fid);
			return (1);
			}
		}
	strcpy(target_buf, filename);
	return (0);							/* can't find one */
	}



/*  getenv.C	*/

/*
	This file contains the routine for searching the environment
	area for a given string.

	Interface:

		getenv(search_str, buffer, buf_len)

		where search_str is the actual string being searched for,
			  buffer is the location to store the string if found,
			  buf_len is the maximum number of characters to store
			          into the buffer (includes terminator).

        -1 is returned if the environment variable isn't found,
        otherwise the length of the string copied is returned.
*/

extern unsigned _pcb;		/* c88 variable with the PSP segment */

int getenv(search_str, buffer, buf_len)
	char *search_str, *buffer; int buf_len; {
	int envseg,envoff,searchoff,len;
	char ch;

	envoff=0;	
	buffer[0]=0;
	envseg=(_peek(0x2d,_pcb) << 8)+_peek(0x2c,_pcb);
	while (ch=_peek(envoff++,envseg)) {
		searchoff=0;
		while (ch == search_str[searchoff++]) ch=_peek(envoff++,envseg);
		if (ch == '=' && search_str[searchoff-1] == 0) {
			len=-1;
			do {
				*buffer++=ch=_peek(envoff++,envseg);
				len++;
				}
			while (ch && buf_len--);
			return len;
			}
		while (ch=_peek(envoff++,envseg));
		}
	return -1;
	}

/*  putenv.C	*/

/*
	This file contains the routine for searching the environment
	area for a given string and replacing it.

	Interface:

		putenv(search_str, buffer)

		where search_str is the actual string being searched for,
			  buffer is the location of new string.
        -1 is returned if the environment variable isn't found,
        otherwise 0 is returned.
*/

int putenv(search_str, buffer)
	char *search_str, *buffer; {
	int  envseg,envoff,searchoff,len,oldlen,newlen,envrep;
	char ch,count,first;

	envoff=0;	
	buffer[0]=0;
	envseg=(_peek(0x2d,_pcb) << 8)+_peek(0x2c,_pcb);
	while (ch=_peek(envoff++,envseg)) {
		searchoff=0;
		while (ch == search_str[searchoff++]) ch=_peek(envoff++,envseg);
		if (ch == '=' && search_str[searchoff-1] == 0) {
			count=0;
			first=1;
			len=2;
			oldlen=0;
			envrep=envoff;
			while (count  < 2) {
				while (_peek(envoff++,envseg)) {
					len++;
					if (first) oldlen++;
					}
				first=0;
				if (_peek(envoff++,envseg) == 0) count++;
				else len++;
				}
			newlen=strlen(buffer);
			_lmove(len,envrep-newlen+oldlen,envseg,envrep,envseg);
			_lmove(newlen,&buffer,_showds(),envrep,envseg);
			return 0;
			}
		while (ch=_peek(envoff++,envseg));
		}
	return -1;
	}
