/*
 *  Released under the GNU LGPL.  See http://www.gnu.org/licenses/lgpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  This library is free software * you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundatation * either version 2.1 of the License, or
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY * without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 *  License for more details.
 */
/*	PRINTF.C  --	FORMATTED OUTPUT ROUTINE. SPRINTF IS HERE ALSO.	*/

#define NULL 0
#define EOF -1		/* Physical EOF returned by low level I/O functions */
#define ERROR -1	/* General "on error" return value */
#define OK 0		/* General purpose "no error" return value */
#define MAXLINE 258	/* Longest line of input expected from the console */
#define TRUE 1		/* general purpose true truth value	*/
#define FALSE 0		/* general purpose false truth value 	*/
char toupper(), isdigit();

/*
	printf

	usage:
		printf(format, arg1, arg2, ...);
	
	Note that since the "_spr" function is used to
	form the output string, and then "puts" is used to
	actually print it out, care must be taken to 
	avoid generating null (zero) bytes in the output,
	since such a byte will terminate printing of the
	string by puts. Thus, a statment such as:

		printf("%c foo",'\0');

	would print nothing at all.

*/

printf(format)
char *format;
{
	char line[MAXLINE];
	_spr(line,&format);	/* use "_spr" to form the output */
	puts(line);		/* and print out the line	 */
}
/*
	sprintf:
	Like fprintf, except a string pointer is specified
	instead of a buffer pointer. The text is written
	directly into memory where the string pointer points.

	Usage:
		sprintf(string,format,arg1, arg2, ...);
*/

sprintf(buffer,format)
char *buffer, *format;
{
	_spr(buffer,&format);	/* call _spr to do all the work */
	return buffer;
}

/*
	General formatted output conversion routine, used by
	fprintf and sprintf..."line" is where the output is
	written, and "fmt" is a pointer to an argument list 
	which must consist of a format string pointer and
	subsequent list of (optional) values. Having arguments
	passed on the stack works out a heck of a lot neater
	than it did before when the args were passed via an
	absolute vector in low memory!
*/

union {long dword; char bytes[8];};

_spr(line,fmt)
char *line, **fmt;
{
	char _uspr();
	char _duspr();
	char c, base, *sptr, *format;
	char wbuf[MAXLINE], *wptr, pf, ljflag, zfflag, plusf, blankf, sharpf;
	int width, precision,  *args, siz;

	format = *fmt++;    /* fmt first points to the format string	*/
	args = fmt;	    /* now fmt points to the first arg value	*/

	while (c = *format++)
	  if (c == '%') {
	    wptr = wbuf;
	    ljflag = pf = zfflag = plusf = blankf = sharpf = 0;
   		precision = 6;

		while(1) {
			switch(c = *format) {
				case '-':	ljflag++;
							format++;
							continue;
		    	case '+':	plusf++;
							format++;
		    				continue;
		    	case ' ':	blankf++;
							format++;
		    				continue;
				case '#':	sharpf++;
							format++;
							continue;
				case '0':	zfflag++;
							format++;
							continue;
				}
			break;
		    }

	    if(isdigit(c))
	    	width = _gv2(&format);
	    else if(c == '*') {
	    	width = *args++;
	    	if(width < 0) {
	    		ljflag++;
	    		width = -width;
	    		}
	    	format++;
	    	}
	    else
	    	width = 0;

	    if ((c = *format++) == '.') {
		    if(isdigit(c = *format))
		    	precision = _gv2(&format);
	    	else if(c == '*') {
		    	if((precision = *args++) < 0)
		    		precision = -precision;
		    	format++;
		    	}
		    pf++;
		    if(precision > MAXLINE - 1)
		    	precision = MAXLINE - 1;
			c=*format++;
			}

		if (c == 'l' || c == 'L') {
		    c = toupper(*format++);
	     	}

	    switch(c) {

		case 'd':	if (*args < 0) {
						*wptr++ = '-';
						*args = -*args;
						width--;
			    		}
			    	else if(blankf) {
			    		*wptr++ = ' ';
			    		width--;
			    		}
			    	else if(plusf) {
			    		*wptr++ = '+';
			    		width--;
			    		}

		case 'u':	base = 10; goto val;

		case 'x':	base = 16;
					if(sharpf) {
						*wptr++ = '0';
						*wptr++ = 'x';
						width -= 2;
						}
					goto val;

		case 'o':	base = 8;  /* note that arbitrary bases can be
				         added easily before this line */
					if(sharpf) {
						*wptr++ = '0';
						width--;
						}
					goto val;

		     val:  width -= _uspr(&wptr,*args++,base);
			   goto pad;

		case 'D':	if (args->dword < 0) {
						*wptr++ = '-';
						args->dword=-args->dword;
						width--;
						}
			    	else if(blankf) {
			    		*wptr++ = ' ';
			    		width--;
			    		}
			    	else if(plusf) {
			    		*wptr++ = '+';
			    		width--;
			    		}

		case 'U':	base = 10; goto dval;

		case 'X':	base = 16;
					if(sharpf) {
						*wptr++ = '0';
						*wptr++ = 'x';
						width -= 2;
						}
					goto dval;

		case 'O':	base = 8;
					if(sharpf) {
						*wptr++ = '0';
						width--;
						}

			dval:	width -= _duspr(&wptr,(args)->dword,base);
					args+=2;
				   goto pad;

		case 'e':
		case 'E':
		case 'f':
		case 'F':
		case 'g':
		case 'G':	if (args->bytes[7] & 0x80) {
						*wptr++='-';
						args->bytes[7]&=0x7f;
						width--;
						}
			    	else if(blankf) {
			    		*wptr++ = ' ';
			    		width--;
			    		}
			    	else if(plusf) {
			    		*wptr++ = '+';
			    		width--;
			    		}
					c=toupper(c);
					if (c == 'E')
						siz = _efmt(wptr,args,precision);
					else if (c == 'F')
						siz = _ffmt(wptr,args,precision);
					else
						siz = _gfmt(wptr,args,precision);
					if(sharpf && precision == 0)
						wptr[siz++] = '.';
					width-=siz;
					wptr[siz]=0;
					args+=4;
					goto pad2;

		case 'c':
		case 'C':  *wptr++ = *args++;
				   width--;
				   goto pad;

		case 's':
		case 'S':  	if (!pf)
						precision = MAXLINE - 1;
			   		sptr = *((char **)args)++;
			   		while (*sptr && precision) {
						*wptr++ = *sptr++;
						precision--;
						width--;
			    		}

		     pad:  *wptr = '\0';
		     pad2:	wptr = wbuf;
				   if (!ljflag) {
				   		if( zfflag 
				   			&& (*wptr == '+' || *wptr == '-' || *wptr == ' '))
				   				*line++ = *wptr++;
						while (width-- > 0)
							*line++ = zfflag ? '0' : ' ';
						}

				   while (*line = *wptr++)
					line++;

				   if (ljflag)
					while (width-- > 0)
						*line++ = ' ';
				   break;

		 default:  *line++ = c;

	     }
	  }
	  else *line++ = c;

	*line = '\0';
}

/*
	Internal routine used by "_spr" to perform ascii-
	to-decimal conversion and update an associated pointer:
*/

int _gv2(sptr)
char **sptr;
{
	int n;
	n = 0;
	while (isdigit(**sptr)) n = 10 * n + *(*sptr)++ - '0';
	return n;
}


/*
	Internal function which converts n into an ASCII
	base `base' representation and places the text
	at the location pointed to by the pointer pointed
	to by `string'. Yes, you read that correctly.
*/

char _uspr(string, n, base)
char **string;
unsigned n;
{
	char length;
	if (n<base) {
		*(*string)++ = (n < 10) ? n + '0' : n + 55;
		return 1;
	}
	length = _uspr(string, n/base, base);
	_uspr(string, n%base, base);
	return length + 1;
}

/*
	Same as above except for longs.	*/

char _duspr(string, n, base)
char **string;
long n;
int  base;
{
	char length;
	if (n >= 0 && n < base) {
		*(*string)++ = (n < 10) ? n + '0' : n + 55;
		return 1;
		}
	if (base == 8) {
		length = _duspr(string, n>>3, base);
		_duspr(string, n & 7, base);
		}
	else if (base == 16) {
		length = _duspr(string, n>>4, base);
		_duspr(string, n & 15, base);
		}
	else {
		if (n < 0) {
			length='0';
			while (n < 0 || n > 1000000000) {
				n-=1000000000;
				length++;
				}
			*(*string)++=length;
			return _duspr(string,n,base)+1;
			}
		length = _duspr(string, n/base, base);
		_duspr(string, n%base, base);
		}
	return length + 1;
	}
