/*
 *  Released under the GNU LGPL.  See http://www.gnu.org/licenses/lgpl.txt
 *
 *  This program is part of the DeSmet C Compiler
 *
 *  This library is free software * you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundatation * either version 2.1 of the License, or
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY * without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 *  License for more details.
 */
/*	SCANF.C  --	FORMATTED INPUT. SSCANF IS HERE ALSO.	*/

#define NULL 0
#define EOF -1		/* Physical EOF returned by low level I/O functions */
#define ERROR -1	/* General "on error" return value */
#define OK 0		/* General purpose "no error" return value */
#define MAXLINE 135	/* Longest line of input expected from the console */
#define TRUE 1		/* general purpose true truth value	*/
#define FALSE 0		/* general purpose false truth value 	*/
#define	NCHARS	256	/* scanset size	*/

union {long dword; char byte;};

char toupper(), isdigit(), *gets();

char _scan_line[MAXLINE], *_scanend;

extern int  _oldfile;



/*
	Internal function to position the character
	pointer argument to the next non white-space
	character in the string:
*/

char _igs(){
	char c;
	while (isspace(c = *_scanend)) _scanend++;
	return (c);
	}


/*
	Internal function to convert character c to value
	in base b , or return ERROR if illegal character for that
	base:
*/

static int _bc(c,b)
char c,b;{
	if (isalpha(c = toupper(c))) c -= 55;
         else  if (isdigit(c))  c -= 0x30;
	 else return ERROR;
	if (c > b-1) return ERROR;
		else return c;
	}

/*
	sets up the scanset table
*/

static char * setup(fmt, tab)
char *fmt, *tab; {
	int b, c, d, t = 1;

	if(*fmt == '^') {
		t--;
		fmt++;
		}

	_setmem(tab, NCHARS, !t);

	if((c = *fmt) == ']' || c == '-') { /* first char is special */
		tab[c] = t;
		fmt++;
		}

	while((c = *fmt++) != ']') {
		if(c == '\0')
			return(NULL); /* unexpected end of format */
		if(c == '-' && (d = *fmt) != ']' && (b = fmt[-2]) < d) {
			_setmem(&tab[b], d - b+1, t);
			fmt++;
			}
		else
			tab[c] = t;
		if(c == '\n')
			tab[0] = t;
		}
	return(fmt);
	}


/*
	General formatted input conversion routine. "line" points
	to a string containing ascii text to be converted, and "fmt"
	points to an argument list consisting of first a format
	string and then a list of pointers to the destination objects.

	Appropriate data is picked up from the text string and stored
	where the pointer arguments point according to the format string.
	See K&R for more info. The field width specification is not
	supported by this version.

	NOTE: the "%s" termination character has been changed
	from "any white space" to the character following the "%s"
	specification in the format string. That is, the call

		sscanf(string, "%s:", &str);

	would ignore leading white space (as is the case with all
	format conversions), and then read in ALL subsequent text
	(including newlines) into the buffer "str" until a COLON
	or null byte is encountered.

*/

int _scn(fmt)
char **fmt;{
	char sf, longf, c, base, n, *sptr, *format;
	int sign, val, **args, curlen, maxlen, flen;
	long dval;
	struct {char lobyte,hibyte;};
	union {double dftype; float ftype;};
	double tfloat;

	format = *fmt++;	/* fmt first points to the format string */
	args = fmt;		/* now it points to the arg list */

	n = 0;
	while (c = *format++)
	{
	   if (isspace(c)) continue;	/* skip white space in format string */
	   if (c != '%')		/* if not %, must match text */
	    {
		if (c != _igs()) return n;
		_scanend++;
	    }
	   else		/* process conversion */
	    {
		sign = 1;
		base = 10;
		sf = longf = 0;
		if ((c = *format++) == '*')
		 {
			sf++;		/* if "*" given, supress assignment */
			c = *format++;
		 }
		curlen=0;
		maxlen=255;		/* assume no maximum length	*/
		if (c >= '0' && c <= '9') {
			maxlen=c-'0';
			while (isdigit(c=*format++)) {
				maxlen*=10;
				maxlen+=c-'0';
				}
			}

		if (toupper(c) == 'L')
			c = toupper(*format++);
		else if (toupper(c) == 'H')
			c = *format++;

		switch (c)
		 {
		   case 'x': base = 16;
			     goto doval;

		   case 'o': base = 8;
			     goto doval;

		   case 'd': if (_igs() == '-') {
					sign = -1;
					_scanend++;
					curlen=1;
				    }

	   doval:  case 'u': val = 0;
			     if (_bc(_igs(),base) == ERROR)
				return n;
			     while ((c = _bc(*_scanend++,base)) != 255 && curlen++ < maxlen)
				val = val * base + c;
			     _scanend--;
				if (!sf) {
					**args++ = val * sign;
					n++;
					}
			     break;

		  case 'X':	base = 16; goto dodval;

		  case 'O':	base = 8;  goto dodval;

		  case 'D': if (_igs() == '-') {
				sign = -1;
				_scanend++;
				curlen=1;
				}

	  dodval:	case 'U':	dval = 0;
					longf=1;
					if (_bc(_igs(),base) == ERROR) return n;
					while ((c = _bc(*_scanend++,base)) != 255 &&
						 curlen++ < maxlen)
							dval = dval * base + c;
					_scanend--;
					if (!sf) {
						(*args++)->dword = dval * sign;
						n++;
						}
					break;


		   case 'f':
		   case 'e':
		   case 'E':


		   case 'F':if (_igs() == '-') {
						sign = -1;
						_scanend++;
						curlen=1;
					    }
					 flen=_finput(_scanend,&tfloat,maxlen-curlen);
					 if (flen == 0) return n;
					 _scanend+=flen;
					 if (sign < 0) tfloat=-tfloat;
					 if (!sf) {
						 if (c >= 'e')
						 	(*args++)->ftype=tfloat;
						 else
						 	(*args++)->dftype=tfloat;
						 n++;
						 }
					 break;


		   case 's':
		   case 'S': _igs();
				     sptr = *args;
				     while (curlen++ < maxlen && (c = *_scanend++)
				     	&& !isspace(c))   {

						if (c != '%' && c == *format) {
							format++;
							break;
							}
						if (!sf)
							*sptr++ = c;
					    }				
					 if (c == 0) _scanend--;
				     if (!sf) {
						 n++;
						 *sptr = '\0';
						 args++;
				    	 }
			    	 break;

		   case '[': {
		   			 char tab[NCHARS];

		   			 if((format = setup(format, tab)) == NULL)
		   			 	return EOF;
				     sptr = *args;
				     while (curlen++ < maxlen && (c = *_scanend++) && tab[c])
							if (!sf)
								*sptr++ = c;
				 	 _scanend--;
				     if (!sf) {
						 n++;
						 *sptr = '\0';
						 args++;
				    	 }
			    	 break;
			    	 }

		   case 'c':
		   case 'C': if (!sf) {
						(*args++)->lobyte=*_scanend;
						n++;
			    		}
				     if (*_scanend) _scanend++;
				     break;

		   default:  return n;
		 }
	    }
	   if (!*_scanend) return n;	/* if end of input string, return */
	}
	return n;
	}


/*
	scanf:
	This one accepts a line of input text from the
	console, and converts the text to the required
	binary or alphanumeric form (see Kernighan &
	Ritchie for a more thorough description):
	Usage:
		scanf(format, ptr1, ptr2, ...);

	Returns number of items matched.

	Since a new line of text must be entered from the
	console each time scanf is called, any unprocessed
	text left over from the last call is lost forever.
	This is a difference between BDS scanf and UNIX
	scanf. Another is that the field width specification
	is not supported here.
*/

int scanf(format)
char *format;{
	static char *scanf_from;
	int numin;

	_scanend=scanf_from;
	if (scanf_from == 0 || _oldfile != 0) {
		if (gets(_scan_line) == 0) return -1;	/* get a line of input from user */
		_scanend=_scan_line;
		}
	numin= _scn(&format);	/* and scan it with "_scn"	 */
	_oldfile=scanf_from=0;
	if (*_scanend == ' ' || _igs()) scanf_from=_scanend;
	return numin;
	}




/*
	sscanf:

	Reads a line of text in from the console and scans it
	for variable values specified in the format string. Uses
	"_scn" for actual conversions; see the comments below in
	the _scn function for more details.

	Usage:
		scanf(format,&arg1,&arg2,...);
*/

int sscanf(line,format)
char *line, *format;{
	_scanend=line;
	return _scn(&format);	/* let _scn do all the work */
	}

